from datetime import date

import polars as pl
from polars import col

from ldnode import logger
from ldnode.loader.schemas.models import DataSource, SyncResult
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.files import STK_PRE_QUOTATION, STK_QUOTATION
from ldnode.loader.datasources.common_plans import plan_rename, plan_correct
from ldnode.loader.datasources.common_plans.limit_exprs import (
    third_board_upper_limit_expr,
    third_board_lower_limit_expr,
    second_board_upper_limit_expr,
    second_board_lower_limit_expr,
    main_board_upper_limit_expr,
    main_board_lower_limit_expr,
    main_board_st_upper_limit_expr,
    main_board_st_lower_limit_expr,
)
from ldnode.loader.datasources.sina import SinaWithTimeoutPatch


def stock_realtime_list_days_expr() -> pl.Expr:
    return ((col("trade_date") - col("list_date")).dt.total_days() + 1).cast(pl.Int16).alias("list_days")


class StkQuotationsDataSource(DataSource):
    def load(self, key: str) -> list[SyncResult]:
        """ 生成股票实时截面数据 """
        try:
            if not STK_PRE_QUOTATION.get_local_path().exists():
                return [SyncResult(key, SyncStatus.FAILED, "数据源不存在")]
            # logger.loader.info(f"开始生成股票实时截面数据")
            quotation = SinaWithTimeoutPatch()
            # 读取股票盘前数据
            pre_quotations = pl.read_parquet(STK_PRE_QUOTATION.get_local_path())
            # 下载行情数据
            quotation_data = quotation.stocks(
                [*map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations['code'],
                )],
                prefix=True
            )
            # 计算股票实时截面数据
            lf = pl.DataFrame(quotation_data).transpose(include_header=True, header_name='code').lazy()
            quotations = (
                lf
                .unnest('column_0')
                .pipe(plan_rename)
                .pipe(plan_correct)
                .drop('name', strict=False)  # 新浪名称不带 -U/-W 后缀，丢弃后由盘前数据提供
                .join(
                    pre_quotations.lazy(),
                    on='code',
                    how='right'
                )
                .with_columns(
                    # # 涨跌幅指标
                    (col('close') / col('pre_close') - 1).alias('pct_chg'),  # 涨跌幅
                    (col("open") / col("pre_close") - 1).alias("open_pct_chg"),  # 开盘涨跌幅
                    (col("high") / col("pre_close") - 1).alias("high_pct_chg"),  # 最高涨跌幅
                    (col("low") / col("pre_close") - 1).alias("low_pct_chg"),  # 最低涨跌幅
                    # 市值相关指标
                    (col("float_share") * col("close")).alias("circ_mv"),  # 流通市值（单位：元）
                    (col("total_share") * col("close")).alias("total_mv"),  # 总市值（单位：元）
                    (col("vol") / col("float_share")).alias("turnover"),  # 换手率
                    # 其他指标
                    (col('adj_factor') * col('close') / col('pre_close')).alias('adj_factor'), # 复权因子
                    (col('list_trade_days') + 1).alias("list_trade_days"), # 上市交易日天数
                    stock_realtime_list_days_expr(),  # 上市自然日天数
                    # 计算涨停价格
                    pl.when(col('code').str.ends_with('BJ')).then(
                        third_board_upper_limit_expr # 北交所
                    ).otherwise(
                        pl.when(col('code').str.starts_with('688') | col('code').str.starts_with('30')).then(
                            second_board_upper_limit_expr # 双创
                        ).otherwise(
                            pl.when(col('name').str.to_uppercase().str.contains('ST')).then(
                                main_board_st_upper_limit_expr # 沪深主板ST
                            ).otherwise(
                                main_board_upper_limit_expr # 沪深主板
                            )
                        )
                    ).alias('upper_limit'),
                    # 计算跌停价格
                    pl.when(col('code').str.ends_with('BJ')).then(
                        third_board_lower_limit_expr # 北交所
                    ).otherwise(
                        pl.when(col('code').str.starts_with('688') | col('code').str.starts_with('30')).then(
                            second_board_lower_limit_expr # 双创
                        ).otherwise(
                            pl.when(col('name').str.to_uppercase().str.contains('ST')).then(
                                main_board_st_lower_limit_expr # 沪深主板ST
                            ).otherwise(
                                main_board_lower_limit_expr # 沪深主板
                            )
                        )
                    ).alias('lower_limit'), 
                )
                .with_columns(
                    # 计算涨跌停标志
                    pl.when(col('close') == col('upper_limit')).then(pl.lit(1)).otherwise(
                        pl.when(col('close') == col('lower_limit')).then(pl.lit(-1)).otherwise(pl.lit(0))
                    ).cast(pl.Int8).alias('limit'),
                )
                # 计算财务指标
                .with_columns(
                    (col('total_mv') / col('npatoshopc_ttm')).alias('pe_ttm'), # 市盈率TTM
                    (col('npatoshopc_ttm') / col('total_mv')).alias('ep_ttm'), # 市盈率TTM倒数
                    (col('total_mv') / col('npadnrpatoshaopc_ttm')).alias('d_pe_ttm'), # 扣非市盈率TTM
                    (col('npadnrpatoshaopc_ttm') / col('total_mv')).alias('d_ep_ttm'), # 扣非市盈率TTM倒数
                    (col('total_mv') / col('tetoshopc')).alias('pb'), # 市净率
                    (col('tetoshopc') / col('total_mv')).alias('bp'), # 市净率倒数
                    (col('total_mv') / col('tetoshopc_wo_gw')).alias('pb_wo_gw'), # 扣除商誉市净率
                    (col('tetoshopc_wo_gw') / col('total_mv')).alias('bp_wo_gw'), # 扣除商誉市净率倒数
                    (col('total_mv') / col('toi')).alias('ps'), # 市销率
                    (col('total_mv') / col('toi_ttm')).alias('ps_ttm'), # 市销率TTM
                    (col('total_mv') / col('ncffoa')).alias('pcf'), # 市现率
                    (col('total_mv') / col('ncffoa_ttm')).alias('pcf_ttm'), # 市现率TTM
                    (col('dy_ttm') / col('total_mv')).alias('dyr'), # 股息率
                )
                # 排除停牌
                .filter(
                    ~((col("open") == 0) & (col("bid1") == 0) & (col("ask1") == 0))
                )
                # 清理无用列
                .select(pl.exclude('now'))
            ).collect()
            # 保存股票实时截面数据
            quotations.write_parquet(STK_QUOTATION.get_local_path())
            return [SyncResult(key, SyncStatus.SUCCESS, date.today().strftime("%Y-%m-%d"))]
        except Exception as e:
            logger.loader.error(f"生成股票实时截面数据失败: {e}")
            return [SyncResult(key, SyncStatus.FAILED, str(e))]



from ldnode.calculator.plans.polars.window_slicer import extract_max_window


def test_extract_max_window_reads_polars_rolling_window_size_keyword():
    computed_factors = [[{
        "factor": "volatility",
        "expr": 'col("pct_chg").rolling_std(window_size=242, min_samples=1, ddof=0).over("code") * (242 ** 0.5)',
        "expr_type": 1,
        "is_intermediate": False,
    }]]

    assert extract_max_window(computed_factors) == 242


def test_extract_max_window_reads_polars_rolling_window_size_positional():
    computed_factors = [[{
        "factor": "amount_20",
        "expr": 'col("amount").rolling_sum(20, min_samples=1).over("code")',
        "expr_type": 1,
        "is_intermediate": False,
    }]]

    assert extract_max_window(computed_factors) == 20


def test_extract_max_window_keeps_stage_dependency_sum():
    computed_factors = [
        [{
            "factor": "volatility",
            "expr": 'col("pct_chg").rolling_std(window_size=242, min_samples=1, ddof=0).over("code")',
            "expr_type": 1,
            "is_intermediate": False,
        }],
        [{
            "factor": "volatility_ma",
            "expr": "ts_mean(volatility, 10)",
            "expr_type": 2,
            "is_intermediate": False,
        }],
    ]

    assert extract_max_window(computed_factors) == 252



from ldnode.loader.files import (
    ETF_DATA,
    ETF_PRE_QUOTATION,
    ETF_QUOTATION,
    FUND_DATA,
    FUND_PRE_QUOTATION,
    FUND_QUOTATION,
)
from ldnode.loader.data_time import DATASET_REGISTRY


def test_fund_loader_uses_fund_backtest_data_index() -> None:
    assert FUND_DATA.key == "backtest_data/fund.parquet"


def test_fund_loader_uses_fund_before_trading_data_index() -> None:
    assert FUND_PRE_QUOTATION.key == "before_trading_data/fund.parquet"


def test_fund_loader_uses_fund_realtime_quotation_index() -> None:
    assert FUND_QUOTATION.key == "fund_quotations.pq"


def test_etf_loader_names_remain_compatible_with_fund_indexes() -> None:
    assert ETF_DATA is FUND_DATA
    assert ETF_PRE_QUOTATION is FUND_PRE_QUOTATION
    assert ETF_QUOTATION is FUND_QUOTATION


def test_fund_snapshot_status_reports_legacy_etf_dataset_name() -> None:
    registry = {dataset_name: file for dataset_name, _, file in DATASET_REGISTRY}

    assert registry["ETF_DATA"] is FUND_DATA
    assert "FUND_DATA" not in registry



from ldnode.calculator.schemas import CB_BUILTIN_FACTORS


def test_cb_builtin_factors_include_new_source_columns() -> None:
    assert "adj_factor" in CB_BUILTIN_FACTORS
    assert "volume" in CB_BUILTIN_FACTORS
    assert "amt" in CB_BUILTIN_FACTORS
    assert "remaining_size" in CB_BUILTIN_FACTORS
    assert "issuance_size" in CB_BUILTIN_FACTORS
    assert "redeem_real_days" in CB_BUILTIN_FACTORS
    assert "redeem_count_days" in CB_BUILTIN_FACTORS
    assert "redeem_total_days" in CB_BUILTIN_FACTORS


def test_cb_builtin_factors_exclude_realtime_cashflow_source_columns() -> None:
    assert "pay_interest_day_list" not in CB_BUILTIN_FACTORS



from datetime import date, time

import polars as pl
import patito as pt

from ldnode import logger
from ldnode.loader.schemas.models import DataSource, SyncResult
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.files import FUND_PRE_QUOTATION, FUND_QUOTATION
from ldnode.loader.datasources.sina import SinaWithTimeoutPatch


class Model(pt.Model):
    trade_date: date # 交易日期
    time: time # 交易时间
    code: str # 标的代码
    fund_type: str # 基金类型
    delist_date: date | None # 退市公告日期
    name: str # 标的名称
    pre_close: float # 前收盘价(元)
    open: float # 开盘价(元)
    high: float # 最高价(元)
    low: float # 最低价(元)
    close: float # 收盘价(元)
    vol: int # 成交量(份)
    amount: float # 成交额(元)
    suspension: bool # 停牌标记

    cash_div: float = pt.Field(dtype=pl.Float32) # 现金分红(元/份)
    stk_div: float = pt.Field(dtype=pl.Float32) # 送股分红
    adj_factor: float # 复权因子
    adj_back: float # 后复权因子
    list_trade_days: int = pt.Field(dtype=pl.Int16) # 上市以来交易日天数
    
    list_days: int = pt.Field(dtype=pl.Int16) # 上市以来自然日天数
    pct_chg: float # 涨跌幅
    open_pct_chg: float # 开盘涨跌幅
    high_pct_chg: float # 最高涨跌幅
    low_pct_chg: float # 最低涨跌幅
    limit: int = pt.Field(dtype=pl.Int8) # 涨跌停标记
    
    last_trade_day: bool # 是否为最后一个交易日


def fund_realtime_list_days_expr() -> pl.Expr:
    return ((pl.col("trade_date") - pl.col("list_date")).dt.total_days() + 1).cast(pl.Int16).alias("list_days")


class ETFQuotationsDataSource(DataSource):
    def load(self, key: str) -> list[SyncResult]:
        """ 加载基金实时截面数据 """
        try:
            logger.loader.info(
                f"[Fund] 实时行情计算开始: output_key={key}, pre_quote_key={FUND_PRE_QUOTATION.key}, "
                f"pre_quote_path={FUND_PRE_QUOTATION.get_local_path()}"
            )
            if not FUND_PRE_QUOTATION.get_local_path().exists():
                logger.loader.warning(
                    f"[Fund] 实时行情计算缺少盘前数据: pre_quote_key={FUND_PRE_QUOTATION.key}, "
                    f"pre_quote_path={FUND_PRE_QUOTATION.get_local_path()}"
                )
                return [SyncResult(key, SyncStatus.FAILED, "数据源不存在")]
            quotation = SinaWithTimeoutPatch()  # 使用新浪数据源
            pre_quotations = pl.scan_parquet(FUND_PRE_QUOTATION.get_local_path())
            pre_quote_schema = pre_quotations.collect_schema()
            logger.loader.info(
                f"[Fund] 盘前数据已读取: pre_quote_key={FUND_PRE_QUOTATION.key}, "
                f"columns={pre_quote_schema.names()}"
            )
            # 下载行情数据
            quotation_data = quotation.stocks(
                [*map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations.select(pl.col('code')).collect().to_series(),
                )],
                prefix=True
            )
            quotations = (
                pl.DataFrame(quotation_data)
                .transpose(include_header=True, header_name='code').lazy()
                .unnest('column_0')
                .rename({
                    'close': 'pre_close',
                    'turnover': 'vol',
                    'volume': 'amount',
                    'date': 'trade_date'
                })
                .with_columns(
                    pl.col('code').str.slice(2) + '.' + pl.col('code').str.slice(0, 2).str.to_uppercase(),
                    pl.col('trade_date').str.to_date(),
                    pl.col('time').str.to_time(),
                    pl.when((pl.col('bid1')==0) | (pl.col('ask1')==0)) # 修正涨跌停
                    .then(pl.col('bid1') + pl.col('ask1'))
                    .otherwise((pl.col('bid1') + pl.col('ask1')) / 2)
                    .alias('close'),
                )
                .with_columns(
                    # 修正无挂单close
                    pl.when(pl.col('close')==0)
                    .then(
                        pl.when(pl.col('now')==0).then(pl.col('pre_close'))# 开盘无挂单修正
                        .otherwise(pl.col('now'))
                    ).otherwise(pl.col('close')).alias('close'),
                )
                .join(
                    pre_quotations.rename({'close': '_pre_close'}), 
                    on='code', how='right'
                )
                .with_columns(
                    # 涨跌幅指标
                    (pl.col('close') / pl.col('pre_close') - 1).alias('pct_chg'),  # 涨跌幅
                    (pl.col("open") / pl.col("pre_close") - 1).alias("open_pct_chg"),  # 开盘涨跌幅
                    (pl.col("high") / pl.col("pre_close") - 1).alias("high_pct_chg"),  # 最高涨跌幅
                    (pl.col("low") / pl.col("pre_close") - 1).alias("low_pct_chg"),  # 最低涨跌幅
                    # 其他指标
                    (pl.col('_pre_close') - pl.col('pre_close')).fill_null(0).cast(pl.Float32).alias('cash_div'), # 现金分红(元/份)
                    pl.lit(0).cast(pl.Float32).alias('stk_div'), # 送股分红(元/份)
                    (pl.col('list_trade_days') + 1).fill_null(1).alias("list_trade_days"), # 上市交易日天数
                    fund_realtime_list_days_expr(),  # 上市自然日天数
                    (pl.col('adj_factor') * pl.col('close') / pl.col('pre_close')).fill_null(1).alias('adj_factor'), # 复权因子
                    (pl.col('_pre_close') / pl.col('pre_close') * pl.col('adj_back')).fill_null(1).alias('adj_back'), # 后复权因子
                    # 停牌标记
                    pl.when(
                        (pl.col("open") == 0) & (pl.col("bid1") == 0) & (pl.col("ask1") == 0)
                    ).then(True).otherwise(False).alias('suspension'),
                    pl.lit(0).cast(pl.Int8).alias('limit'), # 涨跌停标记
                    pl.lit(False).alias('last_trade_day'), # 是否为最后一个交易日
                )
                .select(Model.columns) # 保留所需列
                .collect() # 生成dataframe
                .pipe(Model.validate) # 数据验证
            )
            quotations.write_parquet(FUND_QUOTATION.get_local_path())
            fund_type_counts = quotations.group_by("fund_type").len().to_dicts()
            logger.loader.info(
                f"[Fund] 实时行情计算完成: output_key={FUND_QUOTATION.key}, "
                f"output_path={FUND_QUOTATION.get_local_path()}, rows={quotations.height}, "
                f"fund_type_counts={fund_type_counts}"
            )
            return [SyncResult(key, SyncStatus.SUCCESS, date.today().strftime("%Y-%m-%d"))]
        except Exception as e:
            logger.loader.error(
                f"[Fund] 实时行情计算失败: output_key={key}, pre_quote_key={FUND_PRE_QUOTATION.key}, error={e}"
            )
            return [SyncResult(key, SyncStatus.FAILED, str(e))]



from datetime import date
from typing import Optional

import polars as pl
import polars.selectors as cs

from ldnode import logger
from ldnode.calculator.plans.polars import (
    log_factor_weight_stats,
    planning_mod_rank,
    planning_screen_lf,
    planning_screen_weight,
)
from ldnode.calculator.schemas import (
    CB_BUILTIN_FACTORS,
    CB_SCREEN_RESULT_FIXED_COLUMNS,
    ETF_BUILTIN_FACTORS,
    ETF_SCREEN_RESULT_FIXED_COLUMNS,
    SCREEN_CB_SETTING_RELATIVE_FACTORS,
    SCREEN_COMMON_SETTING_RELATIVE_FACTORS,
    SCREEN_ETF_SETTING_RELATIVE_FACTORS,
    SCREEN_STK_SETTING_RELATIVE_FACTORS,
    STK_BUILTIN_FACTORS,
    STK_SCREEN_RESULT_FIXED_COLUMNS,
    BuiltinScope,
    CBBuiltinScope,
    CBNeutralizationMethod,
    CBProcessMissingDataMethod,
    CBScreenParameters,
    ETFBuiltinScope,
    ETFNeutralizationMethod,
    ETFProcessMissingDataMethod,
    ETFScreenParameters,
    ExcludeConditionGroup,
    NeutralizationMethod,
    ProcessMissingDataMethod,
    ScreenParameters,
    STKBuiltinScope,
    STKNeutralizationMethod,
    STKProcessMissingDataMethod,
    STKScreenParameters,
    HoldWeight,
    gen_cb_filter_exprs,
    gen_etf_filter_exprs,
    gen_stk_filter_exprs,
)


def _gen_screen_result_columns(
    fixed_columns: list[str],
    parameters: ScreenParameters,
    setting_relative_factors: dict[str, list[str]],
    show_all_data: bool = False,
) -> list[str]:
    """生成返回数据列"""
    if show_all_data:
        # 显示所有列，base_columns 在前
        return (fixed_columns, cs.exclude(fixed_columns))
    # 计算要展示的自定义因子列
    customized_factors = []
    for fs in parameters.get("computed_factors", []):
        customized_factors.extend([f["factor"] for f in fs if not f["is_intermediate"]])
    # 计算快捷排除设置关联的因子列（合并通用映射 + 品种专属映射）
    merged_setting_factors = {}
    merged_setting_factors.update(SCREEN_COMMON_SETTING_RELATIVE_FACTORS)
    merged_setting_factors.update(setting_relative_factors)
    exclude_setting_factors: list[str] = []
    for setting, factors in merged_setting_factors.items():
        if parameters.get(setting):
            exclude_setting_factors.extend(factors)
    # 计算要展示的排除因子列
    if filter_factors := parameters.get("filter_factors"):
        exclude_factors = [
            f["factor"] for f in ExcludeConditionGroup.flatten(filter_factors)
        ]
    else:
        exclude_factors = []
    # 计算要展示的打分因子列
    rank_factors = [f["factor"] for f in parameters["rank_factors"]]
    weight_factors = []
    hold_method = parameters["hold_method"]
    if hold_method["hold_weight"] == HoldWeight.FACTOR:
        weight_factors.append(hold_method["hold_weight_factor"])
    # 计算要展示的打分列（命名规则与 score_factors.py 保持一致：有中性化时列名带 _neutralization）
    score_columns = [
        f"_{f['factor']}_neutralization_score" if f.get("neutralization_factor") else f"_{f['factor']}_score"
        for f in parameters["rank_factors"]
    ] + ["score"]
    # 返回列定义
    return cs.by_name(
        list(
            dict.fromkeys(
                fixed_columns
                + customized_factors
                + exclude_setting_factors
                + exclude_factors
                + rank_factors
                + weight_factors
                + score_columns
            )
        )
    )


def _planning_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters: ScreenParameters,
    builtin_factors: list[str],
    BuiltinScopes: type[BuiltinScope],
    ProcessMissingDataMethods: type[ProcessMissingDataMethod],
    NeutralizationMethods: type[NeutralizationMethod],
    filter_exprs: list[pl.Expr],
    fixed_columns: list[str],
    setting_relative_factors: dict[str, list[str]],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
) -> pl.LazyFrame:
    """规划选标计算"""
    # 规划选标截面
    screen_lf = screen_lf.pipe(
        planning_screen_lf,
        parameters=parameters,
        builtin_factors=builtin_factors,
        filter_exprs=filter_exprs,
        BuiltinScopes=BuiltinScopes,
        ProcessMissingDataMethods=ProcessMissingDataMethods,
        NeutralizationMethods=NeutralizationMethods,
        filter_fn=lambda x: x.filter(pl.col("trade_date") == trade_date),
    )
    # 添加权重列
    screen_lf = screen_lf.pipe(planning_screen_weight, parameters["hold_method"])
    # 若阈值不为0 且 存在持仓数据 => 规划计算修正排名
    if parameters["hold_method"]["hold_threshold"] and hold_securities:
        position_lf = pl.LazyFrame(
            {"code": hold_securities}, schema={"code": pl.String}
        ).with_columns(pl.lit(1).alias("volume"))
        screen_lf = screen_lf.pipe(
            planning_mod_rank,
            parameters["hold_method"]["hold_threshold"],
            parameters["hold_method"]["hold_range_type"],
            position_lf,
        )
    # 排除得分null的证券
    if exclude_null_score_security:
        screen_lf = screen_lf.filter(pl.col("score").is_not_null())
    selected_expr = pl.col("rank").is_between(
        parameters["hold_method"]["hold_range_value"]["start"],
        parameters["hold_method"]["hold_range_value"]["end"],
    )
    if parameters["hold_method"]["hold_weight"] == HoldWeight.FACTOR:
        selected_expr = selected_expr & pl.col("_weight").gt(0)

    screen_lf = (
        screen_lf
        # 按照rank排序
        .sort("rank", descending=False, nulls_last=True)
        # 添加选中列
        .with_columns(
            selected_expr.fill_null(False).alias("selected"),
            pl.col("time").cast(pl.String).alias("time"),  # 转换 time 为字符串
        )
        # 生成返回数据列
        .select(
            _gen_screen_result_columns(
                fixed_columns, parameters, setting_relative_factors, show_all_data
            )
        )
        # 调整返回列顺序
        .select(
            cs.by_name([c for c in fixed_columns]),  # fixed_columns 在前
            cs.exclude(fixed_columns),  # 其他列在后
        )
    )
    return screen_lf


def _collect_screen_plans(
    screen_plans: list[pl.LazyFrame],
    parameters_list: list[ScreenParameters],
    task_name: str,
) -> list[pl.DataFrame]:
    results = pl.collect_all(screen_plans)
    for parameters, result in zip(parameters_list, results):
        hold_method = parameters["hold_method"]
        if hold_method["hold_weight"] != HoldWeight.FACTOR:
            continue
        factor = hold_method["hold_weight_factor"]
        logger.screen.info("[FactorWeight] 选标 FACTOR 完成 task=%s factor=%s", task_name, factor)
        if factor in result.columns:
            factor_stats_frame = result.select(factor)
            if "trade_date" in result.columns:
                factor_stats_frame = result.select(["trade_date", factor])
            else:
                factor_stats_frame = factor_stats_frame.with_columns(
                    pl.lit(task_name).alias("trade_date")
                )
            log_factor_weight_stats(factor_stats_frame, task_name, factor)
        selected = result.filter(pl.col("selected"))
        if selected.height == 0:
            logger.screen.error("[FactorWeight] 选标全部无效 task=%s factor=%s", task_name, factor)
            raise ValueError(f"FACTOR 权重因子全部无效: {factor}")
    return results


def _drop_cb_duplicate_amount_display_columns(results: list[pl.DataFrame]) -> list[pl.DataFrame]:
    dropped_count = sum(1 for result in results if "amt" in result.columns)
    if dropped_count:
        logger.screen.info(
            f"[CBScreen] 重复成交额展示列过滤完成 removed_column=amt result_groups={dropped_count}"
        )
    return [result.drop("amt", strict=False) for result in results]


def cb_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters_list: list[CBScreenParameters],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
    builtin_factors: list[str] | None = None,
    fixed_columns: list[str] | None = None,
) -> list[pl.DataFrame]:
    """选债"""
    # 生成选债计划
    screen_plans: list[pl.LazyFrame] = [
        _planning_screen(
            screen_lf=screen_lf,
            trade_date=trade_date,
            parameters=parameters,
            builtin_factors=builtin_factors or CB_BUILTIN_FACTORS,
            BuiltinScopes=CBBuiltinScope,
            ProcessMissingDataMethods=CBProcessMissingDataMethod,
            NeutralizationMethods=CBNeutralizationMethod,
            filter_exprs=gen_cb_filter_exprs(parameters),
            fixed_columns=fixed_columns or CB_SCREEN_RESULT_FIXED_COLUMNS,
            setting_relative_factors=SCREEN_CB_SETTING_RELATIVE_FACTORS,
            exclude_null_score_security=exclude_null_score_security,
            show_all_data=show_all_data,
            hold_securities=hold_securities,
        )
        for parameters in parameters_list
    ]
    # 执行选债计划
    return _drop_cb_duplicate_amount_display_columns(
        _collect_screen_plans(screen_plans, parameters_list, "cb_screen")
    )


def stk_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters_list: list[STKScreenParameters],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
    builtin_factors: list[str] | None = None,
    fixed_columns: list[str] | None = None,
):
    """选股"""
    # 生成选股计划
    screen_plans: list[pl.LazyFrame] = [
        _planning_screen(
            screen_lf=screen_lf,
            trade_date=trade_date,
            parameters=parameters,
            builtin_factors=builtin_factors or STK_BUILTIN_FACTORS,
            BuiltinScopes=STKBuiltinScope,
            ProcessMissingDataMethods=STKProcessMissingDataMethod,
            NeutralizationMethods=STKNeutralizationMethod,
            filter_exprs=gen_stk_filter_exprs(parameters),
            fixed_columns=fixed_columns or STK_SCREEN_RESULT_FIXED_COLUMNS,
            setting_relative_factors=SCREEN_STK_SETTING_RELATIVE_FACTORS,
            exclude_null_score_security=exclude_null_score_security,
            show_all_data=show_all_data,
            hold_securities=hold_securities,
        )
        for parameters in parameters_list
    ]
    # 执行选股计划
    return _collect_screen_plans(screen_plans, parameters_list, "stk_screen")


def etf_screen(
    screen_lf: pl.LazyFrame,
    trade_date: date,
    parameters_list: list[ETFScreenParameters],
    exclude_null_score_security: bool = True,
    show_all_data: bool = False,
    hold_securities: Optional[list[str]] = None,
    builtin_factors: list[str] | None = None,
    fixed_columns: list[str] | None = None,
):
    """选基"""
    # 生成选基计划
    screen_plans: list[pl.LazyFrame] = [
        _planning_screen(
            screen_lf=screen_lf,
            trade_date=trade_date,
            parameters=parameters,
            builtin_factors=builtin_factors or ETF_BUILTIN_FACTORS,
            BuiltinScopes=ETFBuiltinScope,
            ProcessMissingDataMethods=ETFProcessMissingDataMethod,
            NeutralizationMethods=ETFNeutralizationMethod,
            filter_exprs=gen_etf_filter_exprs(parameters),
            fixed_columns=fixed_columns or ETF_SCREEN_RESULT_FIXED_COLUMNS,
            setting_relative_factors=SCREEN_ETF_SETTING_RELATIVE_FACTORS,
            exclude_null_score_security=exclude_null_score_security,
            show_all_data=show_all_data,
            hold_securities=hold_securities,
        )
        for parameters in parameters_list
    ]
    # 执行选基计划
    return _collect_screen_plans(screen_plans, parameters_list, "etf_screen")



from datetime import date

import polars as pl

from ldnode.loader.datasources.cb_quotations import with_cb_quotation_list_days
from ldnode.loader.datasources.etf_quotations import fund_realtime_list_days_expr
from ldnode.loader.datasources.stk_quotations import stock_realtime_list_days_expr


def collect_realtime_list_days(expr: pl.Expr, list_date: date) -> list[int]:
    return (
        pl.DataFrame(
            {
                "trade_date": [date(2026, 6, 17)],
                "list_date": [list_date],
            }
        )
        .lazy()
        .with_columns(expr)
        .collect()["list_days"]
        .to_list()
    )


def test_stock_realtime_list_days_aligns_tail_close_inclusive_days() -> None:
    assert collect_realtime_list_days(stock_realtime_list_days_expr(), date(2026, 6, 17)) == [1]
    assert collect_realtime_list_days(stock_realtime_list_days_expr(), date(2026, 6, 1)) == [17]


def test_fund_realtime_list_days_aligns_tail_close_inclusive_days() -> None:
    assert collect_realtime_list_days(fund_realtime_list_days_expr(), date(2026, 6, 17)) == [1]
    assert collect_realtime_list_days(fund_realtime_list_days_expr(), date(2026, 6, 1)) == [17]


def test_cb_realtime_list_days_aligns_tail_close_inclusive_days() -> None:
    result = (
        pl.DataFrame(
            {
                "trade_date": [date(2026, 6, 17), date(2026, 6, 17)],
                "list_date": [date(2026, 6, 17), date(2026, 6, 1)],
            }
        )
        .lazy()
        .pipe(with_cb_quotation_list_days)
        .collect()
    )

    assert result["list_days"].to_list() == [1, 17]
    assert result.schema["list_days"] == pl.Int16



from datetime import date

import polars as pl

from ldnode.calculator.plans.polars.join_resolver import resolve_joins
from ldnode.calculator.plans.polars.join_rules import JOIN_RULES, SCOPE_TO_FILE
from ldnode.loader.files import STK_FINA_IND


def test_finance_scope_uses_stock_fina_ind_dataset() -> None:
    assert SCOPE_TO_FILE[10] is STK_FINA_IND

    cb_finance_rule = JOIN_RULES[(1, 10)]
    assert cb_finance_rule.column_suffix == "_fin"
    assert cb_finance_rule.main_join_key == "code_stk"
    assert cb_finance_rule.aux_join_key == "code"
    assert cb_finance_rule.time_key == "trade_date"
    assert cb_finance_rule.aux_time_key == "trade_date"

    stk_finance_rule = JOIN_RULES[(2, 10)]
    assert stk_finance_rule.column_suffix == "_fin"
    assert stk_finance_rule.aux_time_key == "trade_date"


def test_cb_finance_join_generates_finance_suffix_columns() -> None:
    main_lf = pl.DataFrame(
        {
            "code": ["CB001", "CB002"],
            "code_stk": ["000001.SZ", "000002.SZ"],
            "trade_date": [date(2024, 4, 10), date(2024, 4, 10)],
            "close": [101.0, 102.0],
        }
    ).lazy()
    finance_lf = pl.DataFrame(
        {
            "code": ["000001.SZ", "000002.SZ"],
            "trade_date": [date(2024, 3, 31), date(2024, 3, 31)],
            "bs_nwc": [1200, 2400],
        }
    ).lazy()

    result_lf, builtin_factors = resolve_joins(
        main_lf,
        strategy_scope=1,
        required_datasets=[
            {
                "security_type": 10,
                "columns": ["bs_nwc"],
                "column_suffix": "_fin",
                "pre_join_dsl": [],
            }
        ],
        builtin_factors=["code", "code_stk", "trade_date", "close"],
        aux_data_overrides={10: finance_lf},
    )

    result = result_lf.collect().sort("code")
    assert "bs_nwc_fin" in result.columns
    assert "bs_nwc_fin" in builtin_factors
    assert result["bs_nwc_fin"].to_list() == [1200, 2400]


def test_join_resolver_aliases_requested_aux_join_key_and_time_key() -> None:
    main_lf = pl.DataFrame(
        {
            "code": ["CB001"],
            "code_stk": ["000001.SZ"],
            "trade_date": [date(2024, 4, 10)],
            "close": [101.0],
        }
    ).lazy()
    stk_lf = pl.DataFrame(
        {
            "code": ["000001.SZ"],
            "trade_date": [date(2024, 4, 10)],
            "factor_a": [12.5],
        }
    ).lazy()

    result_lf, builtin_factors = resolve_joins(
        main_lf,
        strategy_scope=1,
        required_datasets=[
            {
                "security_type": 2,
                "columns": ["code", "trade_date", "factor_a"],
                "column_suffix": "_astk",
                "pre_join_dsl": [],
            }
        ],
        builtin_factors=["code", "code_stk", "trade_date", "close"],
        aux_data_overrides={2: stk_lf},
    )

    result = result_lf.collect()
    assert "factor_a_astk" in result.columns
    assert "factor_a_astk" in builtin_factors
    assert "code_astk" in result.columns
    assert "trade_date_astk" in result.columns
    assert "code_astk" in builtin_factors
    assert "trade_date_astk" in builtin_factors
    assert result["code_astk"].to_list() == ["000001.SZ"]
    assert result["trade_date_astk"].to_list() == [date(2024, 4, 10)]
    assert result["factor_a_astk"].to_list() == [12.5]


def test_cb_stock_nested_finance_join_computes_stock_scope_factor() -> None:
    main_lf = pl.DataFrame(
        {
            "code": ["CB001", "CB002"],
            "code_stk": ["000001.SZ", "000002.SZ"],
            "trade_date": [date(2024, 4, 3), date(2024, 4, 3)],
            "close": [101.0, 102.0],
        }
    ).lazy()
    stock_lf = pl.DataFrame(
        {
            "code": ["000001.SZ", "000001.SZ", "000002.SZ", "000002.SZ"],
            "trade_date": [
                date(2024, 4, 2),
                date(2024, 4, 3),
                date(2024, 4, 2),
                date(2024, 4, 3),
            ],
            "close": [11.0, 12.0, 21.0, 22.0],
        }
    ).lazy()
    finance_lf = pl.DataFrame(
        {
            "code": ["000001.SZ", "000002.SZ"],
            "trade_date": [date(2024, 3, 31), date(2024, 3, 31)],
            "bs_nwc": [100.0, 200.0],
        }
    ).lazy()

    result_lf, builtin_factors = resolve_joins(
        main_lf,
        strategy_scope=1,
        required_datasets=[
            {
                "security_type": 2,
                "columns": ["stock_factor_s"],
                "column_suffix": "_stk",
                "nested_required_datasets": [
                    {
                        "security_type": 10,
                        "columns": ["bs_nwc"],
                        "column_suffix": "_fin",
                        "pre_join_dsl": [],
                    }
                ],
                "scoped_computed_factors": [
                    [
                        {
                            "factor": "stock_factor_s",
                            "expr": "ts_shift(close, 1) + bs_nwc_fin",
                            "expr_type": 2,
                            "is_intermediate": False,
                        }
                    ]
                ],
            }
        ],
        builtin_factors=["code", "code_stk", "trade_date", "close"],
        aux_data_overrides={2: stock_lf, 10: finance_lf},
    )

    result = result_lf.collect().sort("code")
    assert "stock_factor_s_stk" in result.columns
    assert "stock_factor_s_stk" in builtin_factors
    assert result["stock_factor_s_stk"].to_list() == [111.0, 221.0]


def test_nested_join_rejects_unsupported_scope() -> None:
    main_lf = pl.DataFrame(
        {
            "code": ["CB001"],
            "code_stk": ["000001.SZ"],
            "trade_date": [date(2024, 4, 3)],
        }
    ).lazy()
    stock_lf = pl.DataFrame(
        {
            "code": ["000001.SZ"],
            "trade_date": [date(2024, 4, 3)],
        }
    ).lazy()

    try:
        resolve_joins(
            main_lf,
            strategy_scope=1,
            required_datasets=[
                {
                    "security_type": 10,
                    "columns": ["x"],
                    "nested_required_datasets": [
                        {"security_type": 2, "columns": ["close"]}
                    ],
                }
            ],
            builtin_factors=["code", "code_stk", "trade_date"],
            aux_data_overrides={2: stock_lf},
        )
    except ValueError as exc:
        assert "P0 仅支持 CB -> STK -> FIN/FINA" in str(exc)
    else:
        raise AssertionError("unsupported nested scope should fail")



import os
from datetime import date

import polars as pl
import pytest

os.environ.setdefault("ENDPOINT_URL", "http://localhost")
os.environ.setdefault("ENDPOINT_ACCESS_ID", "test")
os.environ.setdefault("ENDPOINT_ACCESS_TOKEN", "test")

from ldnode.loader.datasources.cb_quotations.realtime_derived_plan import (  # noqa: E402
    calculate_cb_redeem_remain_days,
    plan_cb_realtime_derived_factors,
)
from ldnode.loader.datasources.cb_quotations.ytm_plan import cal_ytm  # noqa: E402


def test_plan_cb_realtime_derived_factors_calculates_pure_value_and_redeem_state():
    valuation_date = date(2026, 6, 17)
    frame = pl.DataFrame(
        {
            "code": ["110073.SH"],
            "conv_end_date": [date(2028, 6, 16)],
            "pay_interest_day_list": [[date(2026, 6, 16), date(2027, 6, 16), date(2028, 6, 16)]],
            "interst_list": [[0.5, 0.7, 1.0]],
            "maturity_put_price": [110.0],
            "cb_ytm": [2.0],
            "redeem_price_rate": [1.3],
            "redeem_count_days": [3],
            "redeem_total_days": [5],
            "conv_start_date": [date(2025, 1, 1)],
            "redeem_trade_date_window": [
                [
                    date(2026, 6, 13),
                    date(2026, 6, 14),
                    date(2026, 6, 15),
                    date(2026, 6, 16),
                ]
            ],
            "stk_close_window": [[100.0, 140.0, 140.0, 100.0]],
            "conv_price_window": [[100.0, 100.0, 100.0, 100.0]],
            "call_event_status_window": [[]],
            "call_event_ann_date_window": [[]],
            "call_event_recal_date_window": [[]],
            "call_event_reg_date_window": [[]],
            "call_event_price_window": [[]],
            "call_event_priority_window": [[]],
            "close_stk": [140.0],
            "conv_price": [100.0],
        }
    )

    result = plan_cb_realtime_derived_factors(frame.lazy(), valuation_date).collect()

    left_years = max((date(2028, 6, 16) - valuation_date).days + 1, 0) / 365
    cashflows = [
        (date(2027, 6, 16), 0.7),
        (date(2028, 6, 16), 110.0),
    ]
    expected_pure_value = sum(
        cash / ((1 + 2.0 / 100) ** ((cashflow_date - valuation_date).days / 365))
        for cashflow_date, cash in cashflows
    )

    assert result["left_years"].item() == pytest.approx(left_years)
    assert result["pure_value"].item() == pytest.approx(expected_pure_value)
    assert result["redeem_real_days"].item() == 3
    assert result["redeem_remain_days"].item() == 0
    assert result["is_call"].item() == "已满足强赎条件"


def test_plan_cb_realtime_derived_factors_respects_active_non_call_event():
    valuation_date = date(2026, 6, 17)
    frame = pl.DataFrame(
        {
            "conv_end_date": [date(2028, 6, 16)],
            "pay_interest_day_list": [[date(2026, 6, 16), date(2027, 6, 16), date(2028, 6, 16)]],
            "interst_list": [[0.5, 0.7, 1.0]],
            "maturity_put_price": [110.0],
            "cb_ytm": [2.0],
            "redeem_price_rate": [1.3],
            "redeem_count_days": [3],
            "redeem_total_days": [5],
            "conv_start_date": [date(2025, 1, 1)],
            "redeem_trade_date_window": [[date(2026, 6, 16)]],
            "stk_close_window": [[140.0]],
            "conv_price_window": [[100.0]],
            "call_event_status_window": [["公告不强赎"]],
            "call_event_ann_date_window": [[date(2026, 6, 1)]],
            "call_event_recal_date_window": [[date(2026, 7, 1)]],
            "call_event_reg_date_window": [[None]],
            "call_event_price_window": [[None]],
            "call_event_priority_window": [[10]],
            "close_stk": [140.0],
            "conv_price": [100.0],
        }
    )

    result = plan_cb_realtime_derived_factors(frame.lazy(), valuation_date).collect()

    assert result["is_call"].item() == "公告不强赎"
    assert result["redeem_real_days"].item() == 0
    assert result["redeem_remain_days"].item() is None


def test_plan_cb_realtime_derived_factors_keeps_empty_window_unstarted_bond_null():
    valuation_date = date(2026, 6, 17)
    frame = pl.DataFrame(
        {
            "conv_end_date": [date(2032, 6, 16)],
            "pay_interest_day_list": [[date(2026, 6, 16), date(2027, 6, 16), date(2028, 6, 16)]],
            "interst_list": [[0.5, 0.7, 1.0]],
            "maturity_put_price": [110.0],
            "cb_ytm": [2.0],
            "redeem_price_rate": [1.3],
            "redeem_count_days": [3],
            "redeem_total_days": [5],
            "conv_start_date": [date(2026, 12, 8)],
            "redeem_trade_date_window": [[]],
            "stk_close_window": [[]],
            "conv_price_window": [[]],
            "call_event_status_window": [[]],
            "call_event_ann_date_window": [[]],
            "call_event_recal_date_window": [[]],
            "call_event_reg_date_window": [[]],
            "call_event_price_window": [[]],
            "call_event_priority_window": [[]],
            "close_stk": [None],
            "conv_price": [None],
        }
    )

    result = plan_cb_realtime_derived_factors(frame.lazy(), valuation_date).collect()

    assert result["is_call"].item() == "未到转股期"
    assert result["redeem_real_days"].item() is None
    assert result["redeem_remain_days"].item() is None


def test_calculate_cb_redeem_remain_days_respects_rolling_window():
    assert calculate_cb_redeem_remain_days(
        redeem_flags=[True, False, True, False],
        redeem_count_days=3,
        redeem_total_days=4,
    ) == 2


def test_cal_ytm_uses_pay_interest_day_cashflows():
    ytm = cal_ytm(
        {
            "conv_end_date": date(2028, 6, 16),
            "pay_interest_day_list": [
                date(2026, 6, 16),
                date(2027, 6, 16),
                date(2028, 6, 16),
            ],
            "interst_list": [0.5, 0.7, 110.0],
            "maturity_put_price": 110.0,
            "close": 100.0,
            "call_event_status_window": [],
            "call_event_reg_date_window": [],
            "call_event_price_window": [],
        },
        date(2026, 6, 17),
    )

    assert ytm is not None
    assert ytm > 0


def test_plan_cb_realtime_derived_factors_noops_before_etl_columns_arrive():
    frame = pl.DataFrame({"code": ["110073.SH"], "pure_value": [100.0]})

    result = plan_cb_realtime_derived_factors(frame.lazy(), date(2026, 6, 17)).collect()

    assert result.columns == ["code", "pure_value"]
    assert result["pure_value"].item() == 100.0


def test_plan_cb_realtime_derived_factors_uses_announced_call_maturity_only():
    valuation_date = date(2026, 6, 17)
    frame = pl.DataFrame(
        {
            "code": ["118018.SH", "127018.SZ"],
            "conv_end_date": [date(2028, 8, 17), date(2026, 6, 28)],
            "pay_interest_day_list": [[date(2026, 6, 17), date(2027, 6, 17), date(2028, 6, 17)], [date(2026, 6, 17), date(2027, 6, 17), date(2028, 6, 17)]],
            "interst_list": [[0.3, 0.5, 1.0], [0.3, 0.5, 1.0]],
            "maturity_put_price": [110.0, 119.0],
            "cb_ytm": [6.793, 1.4676],
            "redeem_price_rate": [1.3, 1.3],
            "redeem_count_days": [15, 15],
            "redeem_total_days": [30, 30],
            "conv_start_date": [date(2024, 1, 1), date(2024, 1, 1)],
            "redeem_trade_date_window": [[], []],
            "stk_close_window": [[], []],
            "conv_price_window": [[], []],
            "call_event_status_window": [["公告实施强赎"], ["公告到期赎回"]],
            "call_event_ann_date_window": [[date(2026, 6, 1)], [date(2026, 5, 25)]],
            "call_event_recal_date_window": [[None], [None]],
            "call_event_reg_date_window": [[date(2026, 7, 3)], [date(2026, 6, 26)]],
            "call_event_price_window": [[101.3233], [119.0]],
            "call_event_priority_window": [[40], [20]],
            "close_stk": [None, None],
            "conv_price": [None, None],
        }
    )

    result = plan_cb_realtime_derived_factors(frame.lazy(), valuation_date).collect()
    announced_call = result.filter(pl.col("code") == "118018.SH").row(0, named=True)
    maturity_call = result.filter(pl.col("code") == "127018.SZ").row(0, named=True)

    assert announced_call["conv_end_date"] == date(2026, 7, 3)
    assert announced_call["maturity_put_price"] == pytest.approx(101.3233)
    assert announced_call["left_years"] == pytest.approx(
        ((date(2026, 7, 3) - valuation_date).days + 1) / 365
    )
    assert maturity_call["conv_end_date"] == date(2026, 6, 26)
    assert maturity_call["maturity_put_price"] == pytest.approx(119.0)
    assert maturity_call["left_years"] == pytest.approx(
        ((date(2026, 6, 26) - valuation_date).days + 1) / 365
    )



from ldnode import logger
from ldnode.runner.schemas import BaseContext
from ldnode.runner.events import Events


class ResourceController:
    """ 资源控制器 """
    def __init__(self, context: BaseContext, cpu_resource: int, gpu_resource: int):
        self.context = context
        self.cpu_resource = cpu_resource
        self.gpu_resource = gpu_resource

    async def acquire(self, cpu_usage: int, gpu_usage: int) -> bool:
        """ 获取资源 """
        if self.cpu_resource >= cpu_usage and self.gpu_resource >= gpu_usage:
            before_cpu = self.cpu_resource
            before_gpu = self.gpu_resource
            self.cpu_resource -= cpu_usage
            self.gpu_resource -= gpu_usage
            logger.calculator.info(
                "[ResourceLocal] 锁定本地资源 "
                f"cpu_usage={cpu_usage} gpu_usage={gpu_usage} "
                f"before_cpu={before_cpu} after_cpu={self.cpu_resource} "
                f"before_gpu={before_gpu} after_gpu={self.gpu_resource}"
            )
            await self.context.submit_event("calculator", Events.LOCK_RESOURCE, {
                "cpu_resource": cpu_usage,
                "gpu_resource": gpu_usage,
            })
            return True
        else:
            logger.calculator.warning(
                "[ResourceLocal] 本地资源不足 "
                f"cpu_usage={cpu_usage} gpu_usage={gpu_usage} "
                f"free_cpu={self.cpu_resource} free_gpu={self.gpu_resource}"
            )
            return False
        
    async def release(self, cpu_usage: int, gpu_usage: int) -> None:
        """ 释放资源 """
        before_cpu = self.cpu_resource
        before_gpu = self.gpu_resource
        self.cpu_resource += cpu_usage
        self.gpu_resource += gpu_usage
        logger.calculator.info(
            "[ResourceLocal] 释放本地资源 "
            f"cpu_usage={cpu_usage} gpu_usage={gpu_usage} "
            f"before_cpu={before_cpu} after_cpu={self.cpu_resource} "
            f"before_gpu={before_gpu} after_gpu={self.gpu_resource}"
        )
        await self.context.submit_event("calculator", Events.RELEASE_RESOURCE, {
            "cpu_resource": cpu_usage,
            "gpu_resource": gpu_usage,
        })



from ldnode.loader.files import CB_DATA, CB_PRE_QUOTATION


def test_cb_backtest_data_uses_convertible_bond_path() -> None:
    assert CB_DATA.key == "backtest_data/convertible_bond.parquet"


def test_cb_pre_quotation_uses_convertible_bond_path() -> None:
    assert CB_PRE_QUOTATION.key == "before_trading_data/convertible_bond.parquet"



import math
from datetime import date, datetime
from typing import Any

import polars as pl


CB_REALTIME_DERIVED_REQUIRED_COLUMNS = {
    "cb_ytm",
    "interst_list",
    "maturity_put_price",
    "conv_end_date",
    "redeem_price_rate",
    "redeem_count_days",
    "redeem_total_days",
    "conv_start_date",
    "redeem_trade_date_window",
    "stk_close_window",
    "conv_price_window",
    "call_event_status_window",
    "call_event_ann_date_window",
    "call_event_recal_date_window",
    "call_event_reg_date_window",
    "call_event_price_window",
    "call_event_priority_window",
}

_FORCE_REDEEM_SPECIAL_STATUSES = {
    "公告不强赎",
    "公告提示强赎",
    "公告强赎",
    "已公告强赎",
    "公告实施强赎",
    "公告到期赎回",
    "未到转股期",
}

_SINGLE_CASHFLOW_STATUSES = {
    "公告实施强赎",
    "公告提示强赎",
    "公告到期赎回",
}


def _to_date(value: Any) -> date | None:
    if value is None:
        return None
    if isinstance(value, datetime):
        return value.date()
    if isinstance(value, date):
        return value
    return datetime.fromisoformat(str(value)).date()


def _safe_float(value: Any) -> float | None:
    if value is None:
        return None
    try:
        result = float(value)
    except (TypeError, ValueError):
        return None
    if math.isnan(result):
        return None
    return result


def _safe_int(value: Any) -> int | None:
    if value is None:
        return None
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def calculate_cb_left_years(conv_end_date: Any, valuation_date: date) -> float | None:
    end_date = _to_date(conv_end_date)
    if end_date is None:
        return None
    return max((end_date - valuation_date).days + 1, 0) / 365


def calculate_cb_effective_maturity(row: dict[str, Any]) -> dict[str, Any]:
    conv_end_date = _to_date(row.get("conv_end_date"))
    maturity_put_price = _safe_float(row.get("maturity_put_price"))
    event_statuses = row.get("call_event_status_window") or []
    event_reg_dates = row.get("call_event_reg_date_window") or []
    event_prices = row.get("call_event_price_window") or []

    effective_conv_end_date = conv_end_date
    effective_maturity_put_price = maturity_put_price
    for index, status in enumerate(event_statuses):
        if status not in _SINGLE_CASHFLOW_STATUSES:
            continue
        event_reg_date = _to_date(
            event_reg_dates[index] if index < len(event_reg_dates) else None
        )
        event_price = _safe_float(event_prices[index] if index < len(event_prices) else None)
        if event_reg_date is not None:
            effective_conv_end_date = event_reg_date
        if event_price is not None:
            effective_maturity_put_price = event_price

    return {
        "_cb_effective_conv_end_date": effective_conv_end_date,
        "_cb_effective_maturity_put_price": effective_maturity_put_price,
    }


def calculate_cb_future_cashflows(
    row: dict[str, Any],
    valuation_date: date,
) -> list[tuple[date, float]]:
    effective_maturity = calculate_cb_effective_maturity(row)
    effective_conv_end_date = effective_maturity["_cb_effective_conv_end_date"]
    effective_maturity_put_price = effective_maturity["_cb_effective_maturity_put_price"]
    event_statuses = row.get("call_event_status_window") or []

    if any(status in _SINGLE_CASHFLOW_STATUSES for status in event_statuses):
        if effective_conv_end_date is None or effective_maturity_put_price is None:
            return []
        if effective_conv_end_date <= valuation_date:
            return []
        return [(effective_conv_end_date, effective_maturity_put_price)]

    pay_interest_day_list = row.get("pay_interest_day_list") or []
    interest_list = row.get("interst_list") or []
    maturity_put_price = _safe_float(row.get("maturity_put_price"))
    if not pay_interest_day_list or not interest_list or maturity_put_price is None:
        return []

    cash_list = [_safe_float(value) for value in interest_list]
    if any(value is None for value in cash_list):
        return []
    cash_list[-1] = maturity_put_price

    cashflows: list[tuple[date, float]] = []
    for pay_date_value, cash in zip(pay_interest_day_list, cash_list, strict=False):
        pay_date = _to_date(pay_date_value)
        if pay_date is None or pay_date <= valuation_date:
            continue
        cashflows.append((pay_date, cash))
    return cashflows


def calculate_cb_pure_value(row: dict[str, Any]) -> float | None:
    cb_ytm = _safe_float(row.get("cb_ytm"))
    valuation_date = _to_date(row.get("valuation_date"))
    if cb_ytm is None or valuation_date is None:
        return None

    cashflows = calculate_cb_future_cashflows(row, valuation_date)
    if cashflows:
        return sum(
            cash / ((1 + cb_ytm / 100) ** ((cashflow_date - valuation_date).days / 365))
            for cashflow_date, cash in cashflows
        )

    left_years = _safe_float(row.get("left_years"))
    maturity_put_price = _safe_float(row.get("maturity_put_price"))
    interest_list = row.get("interst_list")
    if left_years is None or maturity_put_price is None or not interest_list:
        return None

    cash_list = [_safe_float(value) for value in interest_list]
    if any(value is None for value in cash_list):
        return None
    count = min(math.ceil(left_years), len(cash_list))
    if count <= 0:
        return None

    cash_list[-1] = maturity_put_price
    remains = [left_years - offset for offset in range(count - 1, -1, -1)]
    return sum(
        cash / ((1 + cb_ytm / 100) ** remain)
        for cash, remain in zip(cash_list[-count:], remains, strict=True)
    )


def _active_force_redeem_event(row: dict[str, Any], valuation_date: date) -> str | None:
    event_statuses = row.get("call_event_status_window") or []
    event_ann_dates = row.get("call_event_ann_date_window") or []
    event_recal_dates = row.get("call_event_recal_date_window") or []
    event_priorities = row.get("call_event_priority_window") or []

    active_status: str | None = None
    active_priority = 0
    for index, status in enumerate(event_statuses):
        ann_date = _to_date(event_ann_dates[index] if index < len(event_ann_dates) else None)
        recal_date = _to_date(
            event_recal_dates[index] if index < len(event_recal_dates) else None
        )
        priority = _safe_int(event_priorities[index] if index < len(event_priorities) else None) or 0
        if ann_date is None or ann_date > valuation_date:
            continue
        if status == "公告不强赎" and recal_date is not None and valuation_date >= recal_date:
            continue
        if priority >= active_priority:
            active_status = status
            active_priority = priority
    return active_status


def _latest_non_call_reset_date(row: dict[str, Any], valuation_date: date) -> date | None:
    event_statuses = row.get("call_event_status_window") or []
    event_recal_dates = row.get("call_event_recal_date_window") or []

    reset_dates: list[date] = []
    for index, status in enumerate(event_statuses):
        if status != "公告不强赎":
            continue
        recal_date = _to_date(
            event_recal_dates[index] if index < len(event_recal_dates) else None
        )
        if recal_date is not None and recal_date <= valuation_date:
            reset_dates.append(recal_date)
    if not reset_dates:
        return None
    return max(reset_dates)


def _force_redeem_flags(row: dict[str, Any], valuation_date: date) -> list[tuple[date, bool]]:
    trade_dates = row.get("redeem_trade_date_window") or []
    stock_closes = row.get("stk_close_window") or []
    conv_prices = row.get("conv_price_window") or []
    redeem_price_rate = _safe_float(row.get("redeem_price_rate"))
    if redeem_price_rate is None:
        return []

    flags: list[tuple[date, bool]] = []
    for index, trade_date_value in enumerate(trade_dates):
        trade_date = _to_date(trade_date_value)
        stock_close = _safe_float(stock_closes[index] if index < len(stock_closes) else None)
        conv_price = _safe_float(conv_prices[index] if index < len(conv_prices) else None)
        if trade_date is None or trade_date > valuation_date:
            continue
        if stock_close is None or conv_price is None:
            continue
        force_redeem_price = round(conv_price * redeem_price_rate, 3)
        flags.append((trade_date, stock_close >= force_redeem_price))

    realtime_stock_close = _safe_float(row.get("close_stk"))
    realtime_conv_price = _safe_float(row.get("conv_price"))
    if realtime_stock_close is not None and realtime_conv_price is not None:
        realtime_flag = realtime_stock_close >= round(realtime_conv_price * redeem_price_rate, 3)
        if flags and flags[-1][0] == valuation_date:
            flags[-1] = (valuation_date, realtime_flag)
        else:
            flags.append((valuation_date, realtime_flag))
    return flags


def calculate_cb_redeem_state(row: dict[str, Any]) -> dict[str, Any]:
    valuation_date = _to_date(row.get("valuation_date"))
    if valuation_date is None:
        return {
            "is_call": None,
            "redeem_real_days": None,
            "redeem_remain_days": None,
        }

    redeem_count_days = _safe_int(row.get("redeem_count_days"))
    redeem_total_days = _safe_int(row.get("redeem_total_days"))
    conv_start_date = _to_date(row.get("conv_start_date"))
    active_status = _active_force_redeem_event(row, valuation_date)
    if conv_start_date is not None and valuation_date < conv_start_date:
        active_status = "未到转股期"
    historical_redeem_window_empty = not (row.get("redeem_trade_date_window") or [])
    flags = _force_redeem_flags(row, valuation_date)

    if active_status in _FORCE_REDEEM_SPECIAL_STATUSES:
        return {
            "is_call": active_status,
            "redeem_real_days": (
                None
                if active_status == "未到转股期" and historical_redeem_window_empty
                else 0
            ),
            "redeem_remain_days": None,
        }

    if redeem_count_days is None or redeem_total_days is None:
        return {
            "is_call": "强赎条款缺失",
            "redeem_real_days": None,
            "redeem_remain_days": None,
        }

    reset_date = _latest_non_call_reset_date(row, valuation_date)
    if conv_start_date is not None:
        flags = [(trade_date, flag) for trade_date, flag in flags if trade_date >= conv_start_date]
    if reset_date is not None:
        flags = [(trade_date, flag) for trade_date, flag in flags if trade_date >= reset_date]
    flags = flags[-redeem_total_days:]
    redeem_real_days = sum(1 for _, flag in flags if flag)
    redeem_remain_days = calculate_cb_redeem_remain_days(
        [flag for _, flag in flags],
        redeem_count_days,
        redeem_total_days,
    )
    if redeem_real_days >= redeem_count_days:
        is_call = "已满足强赎条件"
    else:
        is_call = f"{redeem_real_days}/{redeem_count_days}/{redeem_total_days}"

    return {
        "is_call": is_call,
        "redeem_real_days": redeem_real_days,
        "redeem_remain_days": redeem_remain_days,
    }


def calculate_cb_redeem_remain_days(
    redeem_flags: list[bool],
    redeem_count_days: int,
    redeem_total_days: int,
) -> int:
    for future_days in range(redeem_count_days + 1):
        simulated_flags = (redeem_flags + [True] * future_days)[-redeem_total_days:]
        if sum(1 for flag in simulated_flags if flag) >= redeem_count_days:
            return future_days
    return redeem_count_days


def plan_cb_realtime_derived_factors(
    lf: pl.LazyFrame,
    valuation_date: date,
) -> pl.LazyFrame:
    schema_names = set(lf.collect_schema().names())
    if not CB_REALTIME_DERIVED_REQUIRED_COLUMNS.issubset(schema_names):
        return lf
    if "pay_interest_day_list" not in schema_names:
        lf = lf.with_columns(
            pl.lit(None, dtype=pl.List(pl.Date)).alias("pay_interest_day_list")
        )

    redeem_state_dtype = pl.Struct(
        {
            "is_call": pl.String,
            "redeem_real_days": pl.Int64,
            "redeem_remain_days": pl.Int64,
        }
    )
    effective_maturity_dtype = pl.Struct(
        {
            "_cb_effective_conv_end_date": pl.Date,
            "_cb_effective_maturity_put_price": pl.Float64,
        }
    )
    return (
        lf.with_columns(
            pl.struct(
                [
                    "conv_end_date",
                    "maturity_put_price",
                    "call_event_status_window",
                    "call_event_reg_date_window",
                    "call_event_price_window",
                ]
            )
            .map_elements(calculate_cb_effective_maturity, return_dtype=effective_maturity_dtype)
            .alias("_cb_effective_maturity")
        )
        .unnest("_cb_effective_maturity")
        .with_columns(
            pl.struct(["_cb_effective_conv_end_date"])
            .map_elements(
                lambda row: calculate_cb_left_years(
                    row["_cb_effective_conv_end_date"],
                    valuation_date,
                ),
                return_dtype=pl.Float64,
            )
            .alias("left_years"),
        )
        .with_columns(
            pl.struct(
                [
                    "left_years",
                    "pay_interest_day_list",
                    "interst_list",
                    "conv_end_date",
                    "_cb_effective_maturity_put_price",
                    "cb_ytm",
                    "call_event_status_window",
                    "call_event_reg_date_window",
                    "call_event_price_window",
                ]
            )
            .map_elements(
                lambda row: calculate_cb_pure_value(
                    {
                        "left_years": row["left_years"],
                        "valuation_date": valuation_date,
                        "pay_interest_day_list": row["pay_interest_day_list"],
                        "interst_list": row["interst_list"],
                        "conv_end_date": row["conv_end_date"],
                        "maturity_put_price": row["_cb_effective_maturity_put_price"],
                        "cb_ytm": row["cb_ytm"],
                        "call_event_status_window": row["call_event_status_window"],
                        "call_event_reg_date_window": row["call_event_reg_date_window"],
                        "call_event_price_window": row["call_event_price_window"],
                    }
                ),
                return_dtype=pl.Float64,
            )
            .alias("pure_value"),
        )
        .with_columns(
            pl.col("_cb_effective_conv_end_date").alias("conv_end_date"),
            pl.col("_cb_effective_maturity_put_price").alias("maturity_put_price"),
        )
        .drop("_cb_effective_conv_end_date", "_cb_effective_maturity_put_price")
        .with_columns(pl.lit(valuation_date).alias("valuation_date"))
        .with_columns(
            pl.struct(
                [
                    "valuation_date",
                    "conv_start_date",
                    "redeem_price_rate",
                    "redeem_count_days",
                    "redeem_total_days",
                    "redeem_trade_date_window",
                    "stk_close_window",
                    "conv_price_window",
                    "call_event_status_window",
                    "call_event_ann_date_window",
                    "call_event_recal_date_window",
                    "call_event_priority_window",
                    "close_stk",
                    "conv_price",
                ]
            )
            .map_elements(calculate_cb_redeem_state, return_dtype=redeem_state_dtype)
            .alias("_cb_redeem_state")
        )
        .unnest("_cb_redeem_state")
        .drop("valuation_date")
    )



from datetime import datetime

import polars as pl
from polars import col

from ldnode import logger
from ldnode.loader.schemas.enumerations import SyncStatus
from ldnode.loader.schemas.models import DataSource, SyncResult
from ldnode.loader.files import CB_PRE_QUOTATION, CB_QUOTATION, CB_HIST_DATA, STK_HIST_DATA
from ldnode.loader.datasources.common_plans import plan_rename, plan_correct
from ldnode.loader.datasources.common_plans.limit_exprs import (
    cb_lower_limit_expr, 
    cb_upper_limit_expr,
)
from ldnode.loader.datasources.sina import SinaWithTimeoutPatch
from .ytm_plan import plan_ytm
from .conv_prem_plan import plan_conv_prem
from .history_plan import plan_history
from .option_plan import plan_option
from .realtime_derived_plan import (
    CB_REALTIME_DERIVED_REQUIRED_COLUMNS,
    plan_cb_realtime_derived_factors,
)


def with_cb_quotation_trade_min_unit_columns(lf: pl.LazyFrame) -> pl.LazyFrame:
    return lf.with_columns(
        col("vol").alias("volume"),
        col("amount").alias("amt"),
    )


def with_cb_quotation_size_min_unit_columns(lf: pl.LazyFrame) -> pl.LazyFrame:
    min_currency_unit = pl.lit(100000000)
    return lf.with_columns(
        (col("remain_size") * min_currency_unit).alias("remaining_size"),
        (col("issue_size") * min_currency_unit).alias("issuance_size"),
        (col("remain_cap") * min_currency_unit).alias("remaining_market_value"),
    )


def prepare_cb_pre_quotations_for_realtime(pre_quotations: pl.DataFrame) -> pl.DataFrame:
    realtime_columns_from_sina = [
        "trade_date",
        "date",
        "time",
        "pre_close",
        "open",
        "high",
        "low",
        "close",
        "vol",
        "volume",
        "amount",
        "amt",
        "bid1",
        "ask1",
        "pre_close_stk",
        "open_stk",
        "high_stk",
        "low_stk",
        "close_stk",
        "pct_chg_stk",
        "vol_stk",
        "amount_stk",
    ]
    return (
        pre_quotations
        .rename({"adj_factor": "_pre_adj_factor"})
        .drop(realtime_columns_from_sina, strict=False)
    )


def with_cb_quotation_adj_factor(lf: pl.LazyFrame) -> pl.LazyFrame:
    return lf.with_columns(
        (col("_pre_adj_factor") * col("close") / col("pre_close")).alias("adj_factor")
    )


def with_cb_quotation_list_days(lf: pl.LazyFrame) -> pl.LazyFrame:
    return lf.with_columns(
        ((col("trade_date") - col("list_date")).dt.total_days() + 1).cast(pl.Int16).alias("list_days")
    )


class CBQuotationsDataSource(DataSource):
    def load(self, key: str) -> list[SyncResult]:
        """ 生成可转债实时截面数据 """
        try:
            if not CB_PRE_QUOTATION.get_local_path().exists() or not CB_HIST_DATA.get_local_path().exists() or not STK_HIST_DATA.get_local_path().exists():
                return [SyncResult(key, SyncStatus.FAILED, "数据源不存在")]
            # logger.loader.info(f"开始生成可转债实时截面数据")
            quotation = SinaWithTimeoutPatch()
            today = datetime.now().date()
            # 读取转债盘前数据
            pre_quotations = prepare_cb_pre_quotations_for_realtime(
                pl.read_parquet(CB_PRE_QUOTATION.get_local_path())
            )
            logger.loader.info(
                f"[CBQuotation] 盘前数据已读取: pre_quote_key={CB_PRE_QUOTATION.key}, "
                f"pre_quote_path={CB_PRE_QUOTATION.get_local_path()}, "
                f"columns={pre_quotations.columns}"
            )
            missing_realtime_derived_columns = sorted(
                CB_REALTIME_DERIVED_REQUIRED_COLUMNS - set(pre_quotations.columns)
            )
            logger.loader.info(
                "[CBQuotation] 实时现算原材料检查 "
                f"missing_columns={missing_realtime_derived_columns}"
            )
            missing_pay_calendar_columns = sorted(
                {"pay_interest_day_list"} - set(pre_quotations.columns)
            )
            logger.loader.info(
                "[CBQuotation] 付息日历现金流原材料检查 "
                f"missing_columns={missing_pay_calendar_columns}"
            )
            # 下载行情数据
            ### 计算需要下载的标的代码列表
            codes = [
                *map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations['code'],
                ),
                *map(
                    lambda code: code.split('.')[1].lower() + code.split('.')[0], 
                    pre_quotations['code_stk'],
                ),
            ]
            quotation_data = quotation.stocks(codes, prefix=True)
            # 计算可转债实时截面数据
            _quotations = (
                pl.DataFrame(quotation_data).transpose(include_header=True, header_name='code').lazy()
                .unnest('column_0')
                .pipe(plan_rename)
                .pipe(plan_correct)
            )
            logger.loader.info("[CBQuotation] 转债名称口径采用新浪行情 name source=name_right fallback=name")
            quotations = (
                pre_quotations.lazy()
                .join(
                    _quotations,
                    left_on='code',
                    right_on='code',
                    how='left'
                )
                .with_columns(
                    pl.coalesce(col('name_right').cast(pl.String), col('name').cast(pl.String)).alias('name')
                )
                .drop('name_right')
                .drop('name_stk')
                .join(
                    _quotations,
                    left_on='code_stk',
                    right_on='code',
                    how='left',
                    suffix='_stk',
                )
                .pipe(plan_ytm, today)
                # 排除停牌
                .filter(
                    ~((col("open") == 0) & (col("bid1") == 0) & (col("ask1") == 0))
                )
                .pipe(with_cb_quotation_trade_min_unit_columns)
                .with_columns(
                    (col('close') / col('pre_close') - 1).alias('pct_chg'),  # 转债涨跌幅
                    (col('close_stk') / col('pre_close_stk') - 1).alias('pct_chg_stk'),  # 正股涨跌幅
                    (col("amount") / 10000).alias("amount"),  # 成交额（单位：万）
                    (col("vol") / 10).alias("vol"),  # 成交量（单位：手）
                    (col("amount_stk") / 10000).alias("amount_stk"),  # 成交额（单位：万）
                    (col("vol_stk") / 100).alias("vol_stk"),  # 成交量（单位：手）
                    (100 / col("conv_price") * col("close_stk")).alias("conv_value"),  # 转股价值
                )
                .pipe(with_cb_quotation_adj_factor)
                .pipe(with_cb_quotation_list_days)
                .pipe(plan_cb_realtime_derived_factors, today)
                .with_columns(
                    (col("close") / col("conv_value") - 1).alias("conv_prem"),  # 转股溢价率
                    (col("close") / col("pure_value") - 1).alias("bond_prem"),
                )
                .with_columns(  
                    # 纯债溢价率
                    (col("close") + col("conv_prem") * 100).alias("dblow"),  # 双低值
                    (col("pe_ttm") / col("pre_close_stk") * col("close_stk")).alias("pe_ttm"),
                    (col("pb") / col("pre_close_stk") * col("close_stk")).alias("pb"),
                    (col("ps_ttm") / col("pre_close_stk") * col("close_stk")).alias("ps_ttm"),
                    (col("total_mv") / col("pre_close_stk") * col("close_stk") / 1e4).alias("total_mv"),  # 总市值（单位：亿元）
                    (col("circ_mv") / col("pre_close_stk") * col("close_stk") / 1e4).alias("circ_mv"), # 流通市值（单位：亿元）
                    (col("remain_size") / 100 * col("close")).alias("remain_cap")  # 剩余市值
                )
                .pipe(with_cb_quotation_size_min_unit_columns)
                .with_columns(
                    (col("amount") / col("remain_cap") / 10000).alias("turnover"), # 换手率（注意：amount已被修改）
                    (col("remain_cap") / col("total_mv")).alias("cap_mv_rate"),  # 市占比
                    cb_upper_limit_expr.alias('upper_limit'), # 涨停价格
                    cb_lower_limit_expr.alias('lower_limit') # 跌停价格
                )
                .with_columns(
                    # 计算涨跌停标志
                    pl.when(col('close') == col('upper_limit')).then(pl.lit(1)).otherwise(
                        pl.when(col('close') == col('lower_limit')).then(pl.lit(-1)).otherwise(pl.lit(0))
                    ).cast(pl.Int8).alias('limit')
                )
                .pipe(plan_conv_prem)
                .pipe(plan_history)
                .pipe(plan_option)
            ).collect()
            cb_name_contains_z_count = quotations.select(
                col('name').cast(pl.String).str.contains('Z', literal=True).fill_null(False).sum()
            ).item()
            logger.loader.info(
                f"[CBQuotation] 转债名称口径生成完成 rows={quotations.height} "
                f"name_contains_z_count={cb_name_contains_z_count} "
                f"has_name_right={'name_right' in quotations.columns}"
            )
            logger.loader.info(
                "[CBQuotation] 内置字段兼容列生成完成 "
                "columns=volume,amt,remaining_size,issuance_size,remaining_market_value,adj_factor,list_days "
                f"rows={quotations.height}"
            )
            logger.loader.info(
                "[CBQuotation] 实时现算字段生成完成 "
                f"rows={quotations.height} "
                f"pure_value_nulls={quotations['pure_value'].null_count() if 'pure_value' in quotations.columns else 'missing'} "
                f"is_call_nulls={quotations['is_call'].null_count() if 'is_call' in quotations.columns else 'missing'} "
                f"redeem_real_days_nulls={quotations['redeem_real_days'].null_count() if 'redeem_real_days' in quotations.columns else 'missing'} "
                f"redeem_remain_days_nulls={quotations['redeem_remain_days'].null_count() if 'redeem_remain_days' in quotations.columns else 'missing'}"
            )
            # 对齐 S3 数据范式：S3 上的 cb_quotations.pq 使用 date（Datetime）而非 trade_date（Date），
            # 内部计算管道（plan_history 等）依赖 trade_date 列名，因此在输出前重命名回 date
            if 'trade_date' in quotations.columns:
                quotations = quotations.rename({'trade_date': 'date'})
            # 保存可转债实时截面数据
            quotations.write_parquet(CB_QUOTATION.get_local_path())
            # logger.loader.info(f"生成可转债实时截面数据总耗时: {humanize.naturaldelta(time.time() - start_at)}")
            return [SyncResult(key, SyncStatus.SUCCESS, today.strftime("%Y-%m-%d"))]
        except Exception as e:
            logger.loader.error(f"生成可转债实时截面数据失败: {e}")
            return [SyncResult(key, SyncStatus.FAILED, str(e))]



from datetime import date
from dateutil.relativedelta import relativedelta

import polars as pl
import pandas as pd
from pyxirr import xirr

from .realtime_derived_plan import calculate_cb_future_cashflows


def cal_ytm(row, today: date):
    """ 计算可转债到期收益率YTM """
    try:
        cashflows = calculate_cb_future_cashflows(row, today)
        if cashflows:
            dates_return_cal = [pd.Timestamp(today)] + [
                pd.Timestamp(cashflow_date) for cashflow_date, _ in cashflows
            ]
            value_return_cal = [-row['close']] + [cash for _, cash in cashflows]
        else:
            trade_date = pd.to_datetime(today)
            len_years = len(row['interst_list'])
            dates_return = [
                pd.Timestamp(row['conv_end_date'] + relativedelta(years=-len_years + i + 1))
                for i in range(len_years)
            ]
            value_return = row['interst_list'].copy()
            value_return[-1] = row['maturity_put_price']
            dates_return_pre = [i for i in dates_return if i <= trade_date]
            dates_return_cal = [trade_date] + dates_return[len(dates_return_pre):]
            value_return_cal = [-row['close']] + value_return[len(dates_return_pre):]
        ytm = xirr(dates_return_cal, value_return_cal)
        return ytm
    except Exception:
        return None


def plan_ytm(lf: pl.LazyFrame, today: date) -> pl.LazyFrame:
    """ 计算可转债到期收益率YTM """
    if "pay_interest_day_list" not in set(lf.collect_schema().names()):
        lf = lf.with_columns(
            pl.lit(None, dtype=pl.List(pl.Date)).alias("pay_interest_day_list")
        )
    return (
        lf.with_columns(
            pl.struct([
                'conv_end_date',
                'pay_interest_day_list',
                'interst_list',
                'maturity_put_price',
                'close',
                'call_event_status_window',
                'call_event_reg_date_window',
                'call_event_price_window',
            ])
            .map_elements(lambda row: cal_ytm(row, today), return_dtype=pl.Float64).alias('ytm'),
        )
    )



import os
from datetime import date

import polars as pl

os.environ.setdefault("ENDPOINT_URL", "http://localhost")
os.environ.setdefault("ENDPOINT_ACCESS_ID", "test")
os.environ.setdefault("ENDPOINT_ACCESS_TOKEN", "test")

from ldnode.loader.datasources.cb_quotations import (
    prepare_cb_pre_quotations_for_realtime,
    with_cb_quotation_adj_factor,
    with_cb_quotation_list_days,
    with_cb_quotation_size_min_unit_columns,
    with_cb_quotation_trade_min_unit_columns,
)


def test_cb_quotation_trade_builtin_columns_keep_min_units_before_scaling() -> None:
    result = (
        pl.DataFrame(
            {
                "vol": [12.0],
                "amount": [34.0],
            }
        )
        .lazy()
        .pipe(with_cb_quotation_trade_min_unit_columns)
        .collect()
    )

    assert result["volume"].to_list() == [12.0]
    assert result["amt"].to_list() == [34.0]


def test_cb_quotation_size_builtin_columns_convert_yi_to_min_currency_unit() -> None:
    result = (
        pl.DataFrame(
            {
                "remain_size": [56.0],
                "issue_size": [78.0],
                "remain_cap": [90.0],
            }
        )
        .lazy()
        .pipe(with_cb_quotation_size_min_unit_columns)
        .collect()
    )

    assert result["remaining_size"].to_list() == [5600000000.0]
    assert result["issuance_size"].to_list() == [7800000000.0]
    assert result["remaining_market_value"].to_list() == [9000000000.0]


def test_prepare_cb_pre_quotations_keeps_pre_adj_factor_and_drops_realtime_columns() -> None:
    result = prepare_cb_pre_quotations_for_realtime(
        pl.DataFrame(
            {
                "code": ["111010.SH"],
                "name": ["Z Li"],
                "adj_factor": [2.0],
                "close": [120.0],
                "pre_close": [119.0],
                "close_stk": [30.0],
                "remain_size": [1.0],
            }
        )
    )

    assert "_pre_adj_factor" in result.columns
    assert "adj_factor" not in result.columns
    assert "close" not in result.columns
    assert "pre_close" not in result.columns
    assert "close_stk" not in result.columns
    assert result["remain_size"].to_list() == [1.0]


def test_cb_quotation_adj_factor_uses_stock_realtime_formula() -> None:
    result = (
        pl.DataFrame(
            {
                "_pre_adj_factor": [2.0],
                "close": [121.0],
                "pre_close": [110.0],
            }
        )
        .lazy()
        .pipe(with_cb_quotation_adj_factor)
        .collect()
    )

    assert result["adj_factor"].to_list() == [2.2]


def test_cb_quotation_list_days_matches_backtest_inclusive_natural_days() -> None:
    result = (
        pl.DataFrame(
            {
                "trade_date": [date(2026, 6, 17)],
                "list_date": [date(2026, 6, 1)],
            }
        )
        .lazy()
        .pipe(with_cb_quotation_list_days)
        .collect()
    )

    assert result["list_days"].to_list() == [17]
    assert result.schema["list_days"] == pl.Int16
